# -----------------------------------------------------------------------------
# Description:    General utilities.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import importlib
import copy
import xxhash
import numpy as np
import json
import pathlib
from collections.abc import MutableMapping


def delete_keys_from_dict(dictionary: dict, keys: set):
    """
    Delete keys from dictionary recursively.

    Parameters
    ----------
    dictionary : dict
        to be mutated
    keys : set
        a set of keys to strip from the dictionary
    Returns
    -------
    dict
        a new dictionary that does not included the blacklisted keys
    """
    keys_set = set(keys)  # optimization for the "if key in keys" lookup.

    modified_dict = {}
    for key, value in dictionary.items():
        if key not in keys_set:
            if isinstance(value, MutableMapping):
                modified_dict[key] = delete_keys_from_dict(value, keys_set)
            else:
                modified_dict[key] = value
    return modified_dict


def make_hash(o):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable types (including any lists, tuples, sets, and
    dictionaries).

    from: https://stackoverflow.com/questions/5884066/hashing-a-dictionary
    """

    h = xxhash.xxh64()
    if isinstance(o, (set, tuple, list)):

        return tuple([make_hash(e) for e in o])

    elif isinstance(o, np.ndarray):
        # numpy arrays behave funny for hashing
        h.update(o)
        val = h.intdigest()
        h.reset()
        return val

    elif not isinstance(o, dict):
        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


def import_func_from_string(function_string):
    """
    Based on https://stackoverflow.com/questions/3061/calling-a-function-of-a-module-by-using-its-name-a-string
    """
    mod_name, func_name = function_string.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)
    return func


def load_json_schema(relative_to, filename):
    """
    Load a JSON schema from file. Expects a 'schemas' directory in the same directory as `relative_to`.

    .. tip:: Typical usage of the form `schema = load_json_schema(__file__, 'definition.json')`

    Parameters
    ----------
    relative_to : str
        the file to begin searching from
    filename : str
        the JSON file to load
    Returns
    -------
    dict
        the schema
    """
    path = pathlib.Path(relative_to).resolve().parent.joinpath('schemas', filename)
    with path.open(mode='r') as f:
        return json.load(f)


def without(d, keys):
    """
    Utility that copies a dictionary exluding a specific list of keys.
    """
    if not isinstance(keys, list):
        keys = [keys]
    new_d = d.copy()
    for key in keys:
        new_d.pop(key)
    return new_d


class KeyboardFinish(KeyboardInterrupt):
    """
    Indicates the user has signalled to safely abort/finish the experiment.
    """
    pass



# -----------------------------------------------------------------------------
# Description:    Module containing the MeasurementControl.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import time
import json
import types
from os.path import join

import numpy as np
import adaptive
from qcodes import Instrument
from qcodes import validators as vals
from qcodes.instrument.parameter import ManualParameter, InstrumentRefParameter
from qcodes.utils.helpers import NumpyJSONEncoder
from quantify.data.handling import initialize_dataset, create_exp_folder, snapshot, grow_dataset, trim_dataset
from quantify.measurement.types import Settable, Gettable, is_software_controlled


class MeasurementControl(Instrument):
    """
    Instrument responsible for controlling the data acquisition loop.

    MeasurementControl (MC) is based on the notion that every experiment consists of the following step:

        1. Set some parameter(s)            (settable_pars)
        2. Measure some other parameter(s)  (gettable_pars)
        3. Store the data.

    Example:

        .. code-block:: python

            MC.settables(mw_source1.freq)
            MC.setpoints(np.arange(5e9, 5.2e9, 100e3))
            MC.gettables(pulsar_AQM.signal)
            dataset = MC.run(name='Frequency sweep')


    MC exists to enforce structure on experiments. Enforcing this structure allows:

        - Standardization of data storage.
        - Providing basic real-time visualization.

    MC imposes minimal constraints and allows:

    - Soft loops, experiments in which MC controlled acquisition loop.
    - Hard loops, experiments in which MC is not in control of acquisition.
    - Adaptive loops, setpoints are determined based on measured values.

    """

    def __init__(self, name: str):
        """
        Creates an instance of the Measurement Control.

        Parameters
        ----------
        name : str
            name of this instrument
        """
        super().__init__(name=name)

        # Parameters are attributes that we include in logging and intend the user to change.

        self.add_parameter(
            "verbose",
            docstring="If set to True, prints to std_out during experiments.",
            parameter_class=ManualParameter,
            vals=vals.Bool(),
            initial_value=True,
        )

        self.add_parameter(
            "on_progress_callback",
            vals=vals.Callable(),
            docstring="A callback to communicate progress. This should be a "
            "Callable accepting ints between 0 and 100 indicating percdone.",
            parameter_class=ManualParameter,
            initial_value=None,
        )

        self.add_parameter(
            "soft_avg",
            label="Number of soft averages",
            parameter_class=ManualParameter,
            vals=vals.Ints(1, int(1e8)),
            initial_value=1,
        )

        self.add_parameter(
            'instr_plotmon',
            docstring='Instrument responsible for live plotting. '
            'Can be set to str(None) to disable live plotting.',
            parameter_class=InstrumentRefParameter)

        # TODO add update interval functionality.
        self.add_parameter(
            'update_interval',
            initial_value=0.1,
            docstring=(
                'Interval for updates during the data acquisition loop,' +
                ' everytime more than `update_interval` time has elapsed ' +
                'when acquiring new data points, data is written to file ' +
                'and the live monitoring is updated.'),
            parameter_class=ManualParameter,
            vals=vals.Numbers(min_value=0)
        )

        # variables that are set before the start of any experiment.
        self._settable_pars = []
        self._setpoints = []
        self._gettable_pars = []

        # variables used for book keeping during acquisition loop.
        self._nr_acquired_values = 0
        self._loop_count = 0
        self._begintime = time.time()
        self._last_upd = time.time()

        # variables used for persistence and plotting
        self._dataset = None
        self._exp_folder = None
        self._plotmon_name = ''
        self._plot_info = {'2D-grid': False}

    ############################################
    # Methods used to control the measurements #
    ############################################

    def _reset(self):
        """
        Resets all experiment specific variables for a new run.
        """
        self._nr_acquired_values = 0
        self._loop_count = 0
        self._begintime = time.time()

    def _init(self, name):
        """
        Initializes MC, such as creating the Dataset, experiment folder and such.
        """
        # initialize an empty dataset
        self._dataset = initialize_dataset(self._settable_pars, self._setpoints, self._gettable_pars)

        # cannot add it as a separate (nested) dict so make it flat.
        self._dataset.attrs['name'] = name
        self._dataset.attrs.update(self._plot_info)

        self._exp_folder = create_exp_folder(tuid=self._dataset.attrs['tuid'], name=self._dataset.attrs['name'])
        self._dataset.to_netcdf(join(self._exp_folder, 'dataset.hdf5'))  # Write the empty dataset
        snap = snapshot(update=False, clean=True)  # Save a snapshot of all
        with open(join(self._exp_folder, 'snapshot.json'), 'w') as file:
            json.dump(snap, file, cls=NumpyJSONEncoder, indent=4)

        self._plotmon_name = self.instr_plotmon()
        if self._plotmon_name is not None and self._plotmon_name != '':
            self.instr_plotmon.get_instr().tuid(self._dataset.attrs['tuid'])
            # if the timestamp has changed, this will initialize the monitor
            self.instr_plotmon.get_instr().update()

    def run(self, name: str = ''):
        """
        Starts a data acquisition loop.

        Parameters
        ----------
        name : str
            Name of the measurement. This name is included in the name of the data files.
        Returns
        -------
        :class:`xarray.Dataset`
            the dataset
        """

        self._reset()
        self._init(name)

        self._prepare_settables()

        try:
            if self._is_soft:
                self._run_soft()
            else:
                self._run_hard()
        except KeyboardInterrupt:
            print()
            print("Interrupt signalled, exiting gracefully...")

        self._dataset.to_netcdf(join(self._exp_folder, 'dataset.hdf5'))  # Wrap up experiment and store data
        self._finish()
        self._plot_info = {'2D-grid': False}  # reset the plot info for the next experiment.
        self.soft_avg(1)  # reset software averages back to 1

        return self._dataset

    def run_adaptive(self, name, params):
        """
        Starts a data acquisition loop using an adaptive function.

        .. warning ::
            The functionality of this mode can be complex - it is recommended to read the relevant long form
            documentation.

        Parameters
        ----------
        name : str
            Name of the measurement. This name is included in the name of the data files.
        params : dict
            Key value parameters describe the adaptive function to use, and any further parameters for that function.
        Returns
        -------
        :class:`xarray.Dataset`
            the dataset
        """
        def measure(vec) -> float:
            if len(self._dataset['y0']) == self._nr_acquired_values:
                self._dataset = grow_dataset(self._dataset)

            #  1D sweeps return single values, wrap in a list
            if np.isscalar(vec):
                vec = [vec]

            self._soft_set_and_get(vec, self._nr_acquired_values)
            ret = self._dataset['y0'].values[self._nr_acquired_values]
            self._nr_acquired_values += 1
            self._update("Running adaptively")
            return ret

        def subroutine():
            self._prepare_settables()
            self._prepare_gettable()

            adaptive_function = params.get("adaptive_function")
            af_pars_copy = dict(params)

            # leveraging the adaptive library
            if isinstance(adaptive_function, type) and issubclass(adaptive_function, adaptive.learner.BaseLearner):
                goal = af_pars_copy['goal']
                unusued_pars = ["adaptive_function", "goal"]
                for unusued_par in unusued_pars:
                    af_pars_copy.pop(unusued_par, None)
                learner = adaptive_function(measure, **af_pars_copy)
                adaptive.runner.simple(learner, goal)

            # free function
            if isinstance(adaptive_function, types.FunctionType):
                unused_pars = ["adaptive_function"]
                for unused_par in unused_pars:
                    af_pars_copy.pop(unused_par, None)
                adaptive_function(measure, **af_pars_copy)

        if self.soft_avg() != 1:
            raise ValueError("software averaging not allowed in adaptive loops; currently set to {}."
                             .format(self.soft_avg()))

        self._reset()
        self.setpoints(np.empty((64, len(self._settable_pars))))  # block out some space in the dataset
        self._init(name)
        try:
            subroutine()
        except KeyboardInterrupt:
            print()
            print('Interrupt signalled, exiting gracefully...')

        self._finish()
        self._dataset = trim_dataset(self._dataset)
        self._dataset.to_netcdf(join(self._exp_folder, 'dataset.hdf5'))  # Wrap up experiment and store data
        return self._dataset

    def _run_soft(self):
        while self._get_fracdone() < 1.0:
            self._prepare_gettable()
            for row in self._setpoints:
                self._soft_set_and_get(row, self._curr_setpoint_idx())
                self._nr_acquired_values += 1
                self._update()
            self._loop_count += 1

    def _run_hard(self):
        while self._get_fracdone() < 1.0:
            setpoint_idx = self._curr_setpoint_idx()
            for i, spar in enumerate(self._settable_pars):
                spar.set(self._setpoints[setpoint_idx:, i])
            self._prepare_gettable()

            y_off = 0
            for gpar in self._gettable_pars:
                new_data = gpar.get()  # can return (N, M)
                # if we get a simple array, shape it to (1, M)
                if len(np.shape(new_data)) == 1:
                    new_data = new_data.reshape(1, (len(new_data)))

                for row in new_data:
                    slice_len = setpoint_idx + len(row)  # the slice we will be updating
                    old_vals = self._dataset['y{}'.format(y_off)].values[setpoint_idx:slice_len]
                    old_vals[np.isnan(old_vals)] = 0  # will be full of NaNs on the first iteration, change to 0
                    self._dataset['y{}'.format(y_off)].values[setpoint_idx:slice_len] = self._build_data(row, old_vals)
                    y_off += 1
                self._nr_acquired_values += np.shape(new_data)[1]
            self._update()

    def _build_data(self, new_data, old_data):
        if self.soft_avg() == 1:
            return old_data + new_data
        else:
            return (new_data + old_data * self._loop_count) / (1 + self._loop_count)

    def _soft_set_and_get(self, setpoints: np.ndarray, idx: int):
        """
        Processes one row of setpoints. Sets all settables, gets all gettables, encodes new data in dataset

        .. note ::
            Note: some lines in this function are redundant depending on mode (sweep vs adaptive). Specifically
                - in sweep, the x dimensions are already filled
                - in adaptive, soft_avg is always 1
        """
        # set all individual setparams
        for setpar_idx, (spar, spt) in enumerate(zip(self._settable_pars, setpoints)):
            self._dataset['x{}'.format(setpar_idx)].values[idx] = spt
            spar.set(spt)  # TODO add smartness to avoid setting if unchanged
        # get all data points
        y_offset = 0
        for gpar in self._gettable_pars:
            new_data = gpar.get()
            # if the gettable returned a float, cast to list
            if np.isscalar(new_data):
                new_data = [new_data]
            # iterate through the data list, each element is different y for these x coordinates
            for val in new_data:
                old_val = self._dataset['y{}'.format(y_offset)].values[idx]
                if self.soft_avg() == 1 or np.isnan(old_val):
                    self._dataset['y{}'.format(y_offset)].values[idx] = val
                else:
                    averaged = (val + old_val * self._loop_count) / (1 + self._loop_count)
                    self._dataset['y{}'.format(y_offset)].values[idx] = averaged
                y_offset += 1

    ############################################
    # Methods used to control the measurements #
    ############################################

    def _update(self, print_message: str = None):
        """
        Do any updates to/from external systems, such as saving, plotting, checking for interrupts etc.
        """
        update = time.time() - self._last_upd > self.update_interval() \
            or self._nr_acquired_values == self._max_setpoints
        if update:
            self.print_progress(print_message)
            self._dataset.to_netcdf(join(self._exp_folder, 'dataset.hdf5'))
            if self._plotmon_name is not None and self._plotmon_name != '':
                self.instr_plotmon.get_instr().update()
            self._last_upd = time.time()

    def _prepare_gettable(self):
        """
        Call prepare() on the Gettable, if prepare() exists
        """
        for getpar in self._gettable_pars:
            try:
                getpar.prepare()
            # it's fine if the gettable does not have a prepare function
            except AttributeError:
                pass

    def _prepare_settables(self):
        """
        Call prepare() on all Settable, if prepare() exists
        """
        for setpar in self._settable_pars:
            try:
                setpar.prepare()
            # it's fine if the settable does not have a prepare function
            except AttributeError:
                pass

    def _finish(self):
        """
        Call finish() on all Settables and Gettables, if finish() exists
        """
        for p in self._gettable_pars and self._settable_pars:
            try:
                p.finish()
            # it's fine if the parameter does not have a finish function
            except AttributeError:
                pass

    @property
    def _is_soft(self):
        """
        Whether this MeasurementControl controls data stepping
        """
        if any(is_software_controlled(gpar) for gpar in self._gettable_pars):
            if not all(is_software_controlled(gpar) for gpar in self._gettable_pars):
                raise Exception("Control mismatch; all Gettables must have the same Control Mode")
            return True
        return False

    @property
    def _max_setpoints(self):
        """
        The total number of setpoints to examine
        """
        return len(self._setpoints) * self.soft_avg()

    def _curr_setpoint_idx(self):
        """
        Returns the current position through the sweep
        Updates the _soft_iterations_completed counter as it may have rolled over

        Returns
        -------
        int
            setpoint_idx
        """
        acquired = self._nr_acquired_values
        setpoint_idx = acquired % len(self._setpoints)
        self._loop_count = acquired // len(self._setpoints)
        return setpoint_idx

    def _get_fracdone(self):
        """
        Returns the fraction of the experiment that is completed.
        """
        return self._nr_acquired_values / self._max_setpoints

    def print_progress(self, progress_message: str = None):
        percdone = self._get_fracdone()*100
        elapsed_time = time.time() - self._begintime
        if not progress_message:
            progress_message = (
                "\r {percdone}% completed \telapsed time: "
                "{t_elapsed}s \ttime left: {t_left}s".format(
                    percdone=int(percdone),
                    t_elapsed=round(elapsed_time, 1),
                    t_left=round((100.0 - percdone) / percdone * elapsed_time, 1)
                    if percdone != 0
                    else "",
                )
            )
        if self.on_progress_callback() is not None:
            self.on_progress_callback()(percdone)
        if percdone != 100:
            end_char = ""
        else:
            end_char = "\n"
        if self.verbose():
            print("\r", progress_message, end=end_char)

    ####################################
    # Non-parameter get/set functions  #
    ####################################

    def settables(self, settable_pars):
        """
        Define the settable parameters for the acquisition loop.

        The :class:`~quantify.measurement.Settable` helper class defines the requirements for a Settable object.

        Parameters
        ---------
        settable_pars
            parameter(s) to be set during the acquisition loop, accepts a list or tuple of multiple Settable objects
            or a single Settable object.
        """
        # for native nD compatibility we treat this like a list of settables.
        if not isinstance(settable_pars, (list, tuple)):
            settable_pars = [settable_pars]

        self._settable_pars = []
        for settable in settable_pars:
            self._settable_pars.append(Settable(settable))

    def setpoints(self, setpoints):
        """
        Set setpoints that determine values to be set in acquisition loop.

        .. tip::

            Use :code:`np.colstack((x0, x1))` to reshape multiple
            1D arrays when setting multiple setables.

        Parameters
        ----------
        setpoints : :class:`numpy.ndarray`
            An array that defines the values to loop over in the experiment.
            The shape of the array has to be either (N,) (N,1) for a 1D loop or (N, M) in the case of an MD loop.
        """
        if len(np.shape(setpoints)) == 1:
            setpoints = setpoints.reshape((len(setpoints), 1))
        self._setpoints = setpoints

        # set to False whenever new setpoints are defined.
        # this gets updated after calling setpoints_2D.
        self._plot_info['2D-grid'] = False

    def setpoints_grid(self, setpoints):
        """
        Set a setpoint grid that determine values to be set in the acquisition loop. Updates the setpoints in a grid
        by repeating the setpoints M times and filling the second column with tiled values.

        Parameters
        ----------
        setpoints : list
            The values to loop over in the experiment. The grid is reshaped in this order.
        """
        if len(setpoints) == 2:
            self._plot_info['xlen'] = len(setpoints[0])
            self._plot_info['ylen'] = len(setpoints[1])
            self._plot_info['2D-grid'] = True
        self._setpoints = tile_setpoints_grid(setpoints)

    def gettables(self, gettable_pars):
        """
        Define the parameters to be acquired during the acquisition loop.

        The :class:`~quantify.measurement.Gettable` helper class defines the requirements for a Gettable object.

        Parameters
        ----------
        gettable_pars
            parameter(s) to be get during the acquisition loop, accepts:
                 - list or tuple of multiple Gettable objects
                 - a single Gettable object
        """
        if not isinstance(gettable_pars, (list, tuple)):
            gettable_pars = [gettable_pars]

        self._gettable_pars = []
        for gpar in gettable_pars:
            self._gettable_pars.append(Gettable(gpar))


def tile_setpoints_grid(setpoints):
    """
    Tile setpoints into an n-dimensional grid.

    .. warning ::

        using this method typecasts all values into the same type. This may lead to validator errors when setting
        e.g., a float instead of an int.

    Parameters
    ----------
    setpoints : list(:class:`numpy.ndarray`)
        A list of arrays that defines the values to loop over in the experiment. The grid is reshaped in this order.
    Returns
    -------
    :class:`numpy.ndarray`
        an array with repeated x-values and tiled xn-values.
    """
    xn = setpoints[0].reshape((len(setpoints[0]), 1))
    for setpoints_n in setpoints[1:]:
        curr_l = len(xn)
        new_l = len(setpoints_n)
        col_stack = []
        for i in range(0, np.size(xn, 1)):
            col_stack.append(np.tile(xn[:, i], new_l))
        col_stack.append(np.repeat(setpoints_n, curr_l))
        xn = np.column_stack(col_stack)
    return xn



# -----------------------------------------------------------------------------
# Description:  Helpers for performing experiments.
# Repository:   https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
from qcodes import Instrument
from quantify.data.types import TUID
from quantify.data.handling import load_snapshot
from quantify.visualization.pyqt_plotmon import PlotMonitor_pyqt


def load_settings_onto_instrument(instrument: Instrument, tuid: TUID, datadir: str = None):
    """
    Loads settings from a previous experiment onto a current :class:`~qcodes.instrument.base.Instrument`. This
    information is loaded from the 'snapshot.json' file in the provided experiment directory.

    Parameters
    ----------
    instrument : :class:`~qcodes.instrument.base.Instrument`
        the instrument to be configured.
    tuid : :class:`~quantify.data.types.TUID`
        the TUID of the experiment.
    datadir : str
        path of the data directory. If `None`, uses `get_datadir()` to determine the data directory.
    Raises
    ------
    ValueError
        if the provided instrument has no match in the loaded snapshot.
    """
    instruments = load_snapshot(tuid, datadir)['instruments']
    if instrument.name not in instruments:
        raise ValueError('Instrument "{}" not found in snapshot {}:{}'.format(instrument.name, datadir, tuid))
    for parname, par in instruments[instrument.name]["parameters"].items():
        val = par["value"]
        if val:  # qcodes doesn't like setting to none
            instrument.set(parname, par["value"])


def create_plotmon_from_historical(tuid: TUID):
    """
    Creates a plotmon using the dataset of the provided experiment denoted by the tuid in the datadir.
    Loads the data and draws any required figures.

    Parameters
    ----------
    tuid : :class:`~quantify.data.types.TUID`
        the TUID of the experiment.
    Returns
    -------
    :class:`quantify.visualization.pyqt_plotmon.PlotMonitor_pyqt`
        the plot
    """
    plot = PlotMonitor_pyqt(tuid)
    plot.tuid(tuid)
    plot.update()
    return plot



# -----------------------------------------------------------------------------
# Description:    Module containing the core types for use with the MeasurementControl.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import jsonschema
from quantify.utilities.general import load_json_schema


class Settable:
    """
    Defines the Settable concept, which is considered complete if the given type satisfies the following:
    This object does not wrap the passed in object but simply verifies and returns it.

    .. jsonschema:: schemas/Settable.json#/attrs
    .. jsonschema:: schemas/Settable.json#/methods
    """

    schema = load_json_schema(__file__, 'Settable.json')

    def __new__(cls, obj):
        jsonschema.validate(vars(obj), Settable.schema['attrs'])
        jsonschema.validate(dir(obj), Settable.schema['methods'])
        return obj


class Gettable:
    """
    Defines the Gettable concept, which is considered complete if the given type satisfies the following:
    This object does not wrap the passed in object but simply verifies and returns it.

    .. jsonschema:: schemas/Gettable.json#/attrs
    .. jsonschema:: schemas/Gettable.json#/methods
    """

    schema = load_json_schema(__file__, 'Gettable.json')

    def __new__(cls, obj):
        jsonschema.validate(vars(obj), Gettable.schema['attrs'])
        jsonschema.validate(dir(obj), Gettable.schema['methods'])
        return obj


def is_software_controlled(obj):
    if hasattr(obj, 'soft'):
        return obj.soft
    return True



# -----------------------------------------------------------------------------
# Description:    Module containing the core data concepts of quantify.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import datetime


class TUID(str):
    """
    A human readable unique identifier based on the timestamp.

    A tuid is a string formatted as ``YYYYmmDD-HHMMSS-sss-******``.
    The tuid serves as a unique identifier for experiments in quantify see also :mod:`~quantify.data.handling`.
    """

    def __init__(self, value):
        self.is_valid(value)

    def datetime(self):
        """
        Returns
        -------
        :class:`~python:datetime.datetime`
            object corresponding to the TUID
        """
        return datetime.datetime.strptime(self[:18], '%Y%m%d-%H%M%S-%f')

    def uuid(self):
        """
        Returns
        -------
        str
            the uuid (universally unique identifier) component of the TUID, corresponding to the last 6 characters.
        """
        return self[20:]

    @classmethod
    def is_valid(cls, tuid):
        """
        Test if tuid is valid.
        A valid tuid is a string formatted as ``YYYYmmDD-HHMMSS-sss-******``.

        Parameters
        ----------
        tuid : str
            a tuid string

        Returns
        -------
        bool
            True if the string is a valid TUID.

        Raises
        ------
        ValueError
            Invalid format
        """
        cls.datetime(tuid)  # verify date format
        if len(cls.uuid(tuid)) != 6:
            raise ValueError("Invalid format")

        return True



import pytest
import os
import re
import datetime
import pprint
from pathlib import Path


def test_copyrights():
    skip_files = {'__init__.py'}
    failures = {}
    quantify_path = Path(__file__).resolve().parent / os.path.join("..", "quantify")
    current_year = str(datetime.datetime.now().year)
    for root, dirs, files in os.walk(quantify_path):
        path = os.path.relpath(root, quantify_path).split(os.sep)
        for file in files:
            if file[-3:] == '.py' and file not in skip_files:
                filepath = os.path.join(*path, file)
                with open(os.path.join(quantify_path, filepath), 'r') as f:
                    copyright_found = False
                    for line in f.readlines():
                        if re.match(".*Copyright.*Qblox BV", line):
                            copyright_found = True
                            if current_year not in line:
                                failures[filepath] = "No copyright claim for {}.".format(current_year)
                            break
                    if not copyright_found:
                        failures[filepath] = "No copyright found."
    if failures:
        pytest.fail("Bad copyrights:\n{}".format(pprint.pformat(failures)))



# -----------------------------------------------------------------------------
# Description:    Utilities for handling data.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import pathlib
import os
import sys
import json
from datetime import datetime
from uuid import uuid4
import numpy as np
import xarray as xr
from qcodes import Instrument
from quantify.data.types import TUID
from quantify.utilities.general import delete_keys_from_dict


# this is a pointer to the module object instance itself.
this = sys.modules[__name__]

_default_datadir = pathlib.Path(__file__).parent.parent.parent.absolute() / 'data'

this._datadir = None


def gen_tuid(ts=None):
    """
    Generates a :class:`~quantify.data.types.TUID` based on current time.

    Parameters
    ----------
    ts : :class:`datetime.datetime`
        optional, can be passed to ensure the tuid is based on a specific time.

    Returns
    -------
    :class:`~quantify.data.types.TUID`
        timestamp based uid.
    """
    if ts is None:
        ts = datetime.now()
    # ts gives microsecs by default
    (dt, micro) = ts.strftime('%Y%m%d-%H%M%S-.%f').split('.')
    # this ensures the string is formatted correctly as some systems return 0 for micro
    dt = "%s%03d-" % (dt, int(micro) / 1000)
    # the tuid is composed of the timestamp and a 6 character uuid.
    tuid = TUID(dt + str(uuid4())[:6])

    return tuid


def get_datadir():
    """
    Returns the current data directory.
    The data directory can be changed using :func:`~quantify.data.handling.set_datadir`

    Returns
    -------
    str
        the current data directory
    """

    if this._datadir is None:
        this._datadir = this._default_datadir

    return this._datadir


def set_datadir(datadir: str):
    """
    Sets the data directory.

    Parameters
    ----------
    datadir : str
            path of the data directory. If set to ``None``, resets the datadir to the default datadir (``<top_level>/data``).
    """
    this._datadir = datadir


def _locate_experiment_file(tuid: TUID, datadir: str, name: str) -> str:
    if datadir is None:
        datadir = get_datadir()

    daydir = os.path.join(datadir, tuid[:8])

    # This will raise a file not found error if no data exists on the specified date
    exp_folders = list(filter(lambda x: tuid in x, os.listdir(daydir)))
    if len(exp_folders) == 0:
        print(os.listdir(daydir))
        raise FileNotFoundError("File with tuid: {} was not found.".format(tuid))

    # We assume that the length is 1 as tuid is assumed to be unique
    exp_folder = exp_folders[0]

    return os.path.join(daydir, exp_folder, name)


def load_dataset(tuid: TUID, datadir: str = None) -> xr.Dataset:
    """
    Loads a dataset specified by a tuid.

    .. tip::

        This method also works when specifying only the first part of a :class:`~quantify.data.types.TUID`.

    .. note::

        This method uses :func:`xarray.load_dataset` to ensure the file is closed after loading as datasets are
        intended to be immutable after performing the initial experiment.

    Parameters
    ----------
    tuid : str
        a :class:`~quantify.data.types.TUID` string. It is also possible to specify only the first part of a tuid.
    datadir : str
        path of the data directory. If `None`, uses `get_datadir()` to determine the data directory.
    Returns
    -------
    :class:`xarray.Dataset`
        The dataset.
    Raises
    ------
    FileNotFoundError
        No data found for specified date.
    """
    return xr.load_dataset(_locate_experiment_file(tuid, datadir, 'dataset.hdf5'))


def load_snapshot(tuid: TUID, datadir: str = None, file: str = 'snapshot.json') -> dict:
    """
    Loads a snapshot specified by a tuid.

    Parameters
    ----------
    tuid : str
        a :class:`~quantify.data.types.TUID` string. It is also possible to specify only the first part of a tuid.
    datadir : str
        path of the data directory. If `None`, uses `get_datadir()` to determine the data directory.
    file : str
        filename to load
    Returns
    -------
    dict
        The snapshot.
    Raises
    ------
    FileNotFoundError
        No data found for specified date.
    """
    with open(_locate_experiment_file(tuid, datadir, file)) as snap:
        return json.load(snap)


def create_exp_folder(tuid: TUID, name: str = '', datadir=None):
    """
    Creates an empty folder to store an experiment container.

    If the folder already exists, simple return the experiment folder corresponding to the
    :class:`~quantify.data.types.TUID`.

    Parameters
    ----------
    tuid : :class:`~quantify.data.types.TUID`
        a timestamp based human-readable unique identifier.
    name : str
        optional name to identify the folder
    datadir : str
        path of the data directory. If ``None``, uses ``get_datadir()`` to determine the data directory.
    Returns
    -------
    str
        full path of the experiment folder following format: ``/datadir/YYYYmmDD/YYYYmmDD-HHMMSS-sss-******-name/``.
    """
    TUID.is_valid(tuid)

    if datadir is None:
        datadir = get_datadir()
    exp_folder = os.path.join(datadir, tuid[:8], tuid)
    if name != '':
        exp_folder += '-' + name

    os.makedirs(exp_folder, exist_ok=True)
    return exp_folder


def initialize_dataset(settable_pars, setpoints, gettable_pars):
    """
    Initialize an empty dataset based on settable_pars, setpoints and gettable_pars

    Parameters
    ----------
    settable_pars : list
        a list of M settables
    setpoints : :class:`numpy.ndarray`
        an (N*M) array
    gettable_pars : list
        a list of gettables
    Returns
    -------
    :class:`xarray.Dataset`
        the dataset
    """
    darrs = []
    for i, setpar in enumerate(settable_pars):
        darrs.append(xr.DataArray(
            data=setpoints[:, i],
            name='x{}'.format(i),
            attrs={'name': setpar.name, 'long_name': setpar.label, 'unit': setpar.unit})
        )

    numpoints = len(setpoints[:, 0])
    j = 0
    for getpar in gettable_pars:
        #  it's possible for one Gettable to return multiple axes. to handle this, zip the axis info together
        #  so we can iterate through when defining the axis in the dataset
        if not isinstance(getpar.name, list):
            itrbl = zip([getpar.name], [getpar.label], [getpar.unit])
        else:
            itrbl = zip(getpar.name, getpar.label, getpar.unit)

        count = 0
        for idx, info in enumerate(itrbl):
            empty_arr = np.empty(numpoints)
            empty_arr[:] = np.nan
            darrs.append(xr.DataArray(
                data=empty_arr,
                name='y{}'.format(j + idx),
                attrs={'name': info[0], 'long_name': info[1], 'unit': info[2]})
            )
            count += 1
        j += count

    dataset = xr.merge(darrs)
    dataset.attrs['tuid'] = gen_tuid()
    return dataset


def grow_dataset(dataset: xr.Dataset):
    """
    Resizes the dataset by doubling the current length of all arrays.

    Parameters
    ----------
    dataset ::class:`xarray.Dataset`
        the dataset to resize.
    Returns
    -------
    :class:`xarray.Dataset`
        The resized dataset
    """
    darrs = []
    for col in dataset:
        data = dataset[col].values
        darrs.append(xr.DataArray(
            name=dataset[col].name,
            data=np.pad(data, (0, len(data)), 'constant', constant_values=np.nan),
            attrs=dataset[col].attrs
        ))
    dataset = dataset.drop_dims(['dim_0'])
    new_data = xr.merge(darrs)
    return dataset.merge(new_data)


def trim_dataset(dataset: xr.Dataset):
    """
    Trim NaNs from a dataset, useful in the case of a dynamically resized dataset (eg. adaptive loops).

    Parameters
    ----------
    dataset : :class:`xarray.Dataset`
        the dataset to trim.
    Returns
    -------
    :class:`xarray.Dataset`
        The dataset, trimmed and resized if necessary or unchanged.
    """
    for i, val in enumerate(reversed(dataset['y0'].values)):
        if not np.isnan(val):
            finish_idx = len(dataset['y0'].values) - i
            darrs = []
            for col in dataset:
                data = dataset[col].values[:finish_idx]
                darrs.append(xr.DataArray(
                    name=dataset[col].name,
                    data=data,
                    attrs=dataset[col].attrs
                ))
            dataset = dataset.drop_dims(['dim_0'])
            new_data = xr.merge(darrs)
            return dataset.merge(new_data)
    return dataset


########################################################################


def get_latest_tuid(contains: str = '') -> TUID:
    """
    Returns the most recent tuid.

    .. tip::

        This function is similar to :func:`~get_tuids_containing` but is preferred if one is only interested in the
        most recent :class:`~quantify.data.types.TUID` for performance reasons.

    Parameters
    ----------
    contains : str
        an optional string contained in the experiment name.
    Returns
    -------
    :class:`~quantify.data.types.TUID`
        the latest TUID
    Raises
    ------
    FileNotFoundError
        No data found
    """
    return get_tuids_containing(contains, 1)[0]


def get_tuids_containing(contains: str, max_results: int = sys.maxsize) -> list:
    """
    Returns a list of tuids containing a specific label

    .. tip::

        If one is only interested in the most recent :class:`~quantify.data.types.TUID`,
        :func:`~get_latest_tuid` is preferred for performance reasons.

    Parameters
    ----------
    contains : str
        a string contained in the experiment name.
    max_results : int
        maximum number of results to return. Defaults to unlimited.
    Returns
    -------
    list
        A list of :class:`~quantify.data.types.TUID`: objects
    Raises
    ------
    FileNotFoundError
        No data found
    """
    datadir = get_datadir()
    daydirs = list(filter(lambda x: (x.isdigit() and len(x) == 8), os.listdir(datadir)))
    daydirs.sort(reverse=True)
    if len(daydirs) == 0:
        raise FileNotFoundError('There are no valid day directories in the data folder "{}".'.format(datadir))
    tuids = []
    for dd in daydirs:
        expdirs = list(filter(lambda x: (len(x) > 25 and TUID.is_valid(x[:26]) and contains in x),
                              os.listdir(os.path.join(datadir, dd))))
        expdirs.sort(reverse=True)
        for expname in expdirs:
            # Check for inconsistent folder structure for datasets portability
            if dd != expname[:8]:
                raise FileNotFoundError('Experiment container "{}" is in wrong day directory "{}" '.format(expname, dd))
            tuids.append(TUID(expname[:26]))
            if len(tuids) == max_results:
                return tuids
    if len(tuids) == 0:
        raise FileNotFoundError('No experiment found containing "{}"'.format(contains))
    return tuids


def snapshot(update: bool = False, clean: bool = True) -> dict:
    """
    State of all instruments setup as a JSON-compatible dictionary (everything that the custom JSON encoder class
    :class:`qcodes.utils.helpers.NumpyJSONEncoder` supports).

    Parameters
    ----------
    update : bool
        if True, first gets all values before filling the snapshot.
    clean : bool
        if True, removes certain keys from the snapshot to create a more readable and compact snapshot.
    """

    snap = {
        'instruments': {},
        'parameters': {},
    }
    for ins_name, ins_ref in Instrument._all_instruments.items():
        snap['instruments'][ins_name] = ins_ref().snapshot(update=update)

    if clean:
        exclude_keys = {
            "inter_delay",
            "post_delay",
            "vals",
            "instrument",
            "submodules",
            "functions",
            "__class__",
            "raw_value",
            "instrument_name",
            "full_name",
            "val_mapping",
        }
        snap = delete_keys_from_dict(snap, exclude_keys)

    return snap



import pytest
from jsonschema import ValidationError
from qcodes import ManualParameter, Parameter
from quantify.measurement.types import Settable, Gettable


def test_settable():
    x = 5
    with pytest.raises(TypeError):
        Settable(x)

    def test_func(x):
        return 5
    with pytest.raises(ValidationError):
        Settable(test_func)

    manpar = ManualParameter('x')
    Settable(manpar)

    del manpar.unit
    with pytest.raises(ValidationError):
        Settable(manpar)

    std_par = Parameter('x', set_cmd=test_func)
    Settable(std_par)

    class NoName:
        def set(self):
            return
        unit = 'no_name'

    with pytest.raises(ValidationError):
        Settable(NoName())


def test_gettable():
    x = 5
    with pytest.raises(TypeError):
        Gettable(x)

    def test_func():
        return 5
    with pytest.raises(ValidationError):
        Gettable(test_func)

    manpar = ManualParameter('x')
    Gettable(manpar)

    del manpar.unit
    with pytest.raises(ValidationError):
        Gettable(manpar)

    std_par = Parameter('x', get_cmd=test_func)
    Gettable(std_par)



import numpy as np
from quantify.utilities.experiment_helpers import create_plotmon_from_historical
from quantify.data.handling import set_datadir
from tests.helpers import get_test_data_dir


test_datadir = get_test_data_dir()


def test_create_plotmon_from_historical():
    set_datadir(test_datadir)
    plotmon = create_plotmon_from_historical('20200504-191556-002-4209ee')

    x = plotmon.curves[0]['config']['x']
    y = plotmon.curves[0]['config']['y']

    x_exp = np.linspace(0, 5, 50)
    y_exp = np.cos(np.pi * x_exp) * -1
    np.testing.assert_allclose(x[:50], x_exp)
    np.testing.assert_allclose(y[:50], y_exp)

    cfg = plotmon.secondary_QtPlot.traces[0]['config']
    assert np.shape(cfg['z']) == (11, 50)
    assert cfg['xlabel'] == 'Time'
    assert cfg['xunit'] == 's'
    assert cfg['ylabel'] == 'Amplitude'
    assert cfg['yunit'] == 'V'
    assert cfg['zlabel'] == 'Signal level'
    assert cfg['zunit'] == 'V'

    plotmon.close()
    set_datadir(None)



import os
import shutil
import pytest
import xarray as xr
import numpy as np
from quantify.data.types import TUID
import quantify.data.handling as dh
from quantify.measurement.control import MeasurementControl
from datetime import datetime
from qcodes import ManualParameter
import pathlib
from tests.helpers import get_test_data_dir


test_datadir = get_test_data_dir()


def test_gen_tuid():
    ts = datetime.now()
    tuid = dh.gen_tuid(ts)

    assert TUID.is_valid(tuid)
    assert isinstance(tuid, str)
    assert isinstance(tuid, TUID)


def test_initialize_dataset():
    setpar = ManualParameter('x', unit='m', label='X position')
    getpar = ManualParameter('y', unit='V', label='Signal amplitude')
    setable_pars = [setpar]
    setpoints = np.arange(0, 100, 32)
    setpoints = setpoints.reshape((len(setpoints), 1))

    getable_pars = [getpar]
    dataset = dh.initialize_dataset(setable_pars, setpoints, getable_pars)

    assert isinstance(dataset, xr.Dataset)
    assert len(dataset.data_vars) == 2
    assert dataset.attrs.keys() == {'tuid'}
    assert dataset.variables.keys() == {'x0', 'y0'}

    x0 = dataset['x0']
    assert isinstance(x0, xr.DataArray)
    assert x0.attrs['unit'] == 'm'
    assert x0.attrs['name'] == 'x'
    assert x0.attrs['long_name'] == 'X position'

    y0 = dataset['y0']
    assert isinstance(y0, xr.DataArray)
    assert y0.attrs['unit'] == 'V'
    assert y0.attrs['name'] == 'y'
    assert y0.attrs['long_name'] == 'Signal amplitude'


def test_initialize_dataset_2D():
    xpar = ManualParameter('x', unit='m', label='X position')
    ypar = ManualParameter('y', unit='m', label='Y position')
    getpar = ManualParameter('z', unit='V', label='Signal amplitude')
    setable_pars = [xpar, ypar]
    setpoints = np.arange(0, 100, 32)
    setpoints = setpoints.reshape((len(setpoints) // 2, 2))
    getable_pars = [getpar]

    dataset = dh.initialize_dataset(setable_pars, setpoints, getable_pars)

    assert isinstance(dataset, xr.Dataset)
    assert len(dataset.data_vars) == 3
    assert dataset.attrs.keys() == {'tuid'}
    assert set(dataset.variables.keys()) == {'x0', 'x1', 'y0'}


def test_getset_datadir():
    # here to ensure we always start with default datadir
    dh.set_datadir(None)

    default_datadir = dh.get_datadir()
    dd = os.path.split(default_datadir)
    assert dd[-1] == 'data'
    top_level = pathlib.Path(__file__).parent.parent.parent.resolve().name
    assert os.path.split(dd[-2])[-1] == top_level

    dh.set_datadir('my_ddir')
    assert dh.get_datadir() == 'my_ddir'

    # Test resetting to default
    dh.set_datadir(None)
    assert dh.get_datadir() == default_datadir


def test_load_dataset():
    dh.set_datadir(test_datadir)
    tuid = '20200430-170837-001-315f36'
    dataset = dh.load_dataset(tuid=tuid)
    assert dataset.attrs['tuid'] == tuid

    tuid_short = '20200430-170837'
    dataset = dh.load_dataset(tuid=tuid_short)
    assert dataset.attrs['tuid'] == tuid

    with pytest.raises(FileNotFoundError):
        tuid = '20200430-170837-001-3b5f36'
        dh.load_dataset(tuid=tuid)

    with pytest.raises(FileNotFoundError):
        tuid = '20200230-001-170837'
        dh.load_dataset(tuid=tuid)


def test_get_latest_tuid_invalid_datadir():
    dh.set_datadir('some_invalid_datadir')
    with pytest.raises(FileNotFoundError):
        dh.get_latest_tuid()


def test_get_latest_tuid_empty_datadir():
    valid_dir_but_no_data = get_test_data_dir() / 'empty'
    dh.set_datadir(valid_dir_but_no_data)
    with pytest.raises(FileNotFoundError) as excinfo:
        dh.get_latest_tuid()
    assert "There are no valid day directories" in str(excinfo.value)


def test_get_latest_tuid_no_match():
    dh.set_datadir(test_datadir)
    with pytest.raises(FileNotFoundError) as excinfo:
        dh.get_latest_tuid(contains='nonexisting_label')
    assert "No experiment found containing" in str(excinfo.value)


def test_get_latest_tuid_correct_tuid():
    dh.set_datadir(test_datadir)
    tuid = dh.get_latest_tuid(contains='36-Cosine')
    exp_tuid = '20200430-170837-001-315f36'
    assert tuid == exp_tuid


def test_get_tuid_contains():
    dh.set_datadir(test_datadir)
    tuids = dh.get_tuids_containing('Cosine test')
    assert len(tuids) == 2
    assert tuids[0] == '20200504-191556-002-4209ee'
    assert tuids[1] == '20200430-170837-001-315f36'


def test_misplaced_exp_container():
    """
    Ensures user is warned if a dataset was misplaced
    """
    tmp_data_path = os.path.join(
        test_datadir,
        'misplaced_exp_container',
    )
    date = '20201006'
    container = '20201008-191556-002-4209eg-Experiment from my colleague'
    os.makedirs(os.path.join(tmp_data_path, date, container), exist_ok=True)
    dh.set_datadir(tmp_data_path)
    with pytest.raises(FileNotFoundError):
        dh.get_tuids_containing(contains="colleague")

    # cleanup
    shutil.rmtree(tmp_data_path)


def test_snapshot():
    empty_snap = dh.snapshot()
    assert empty_snap == {'instruments': {}, 'parameters': {}}
    test_MC = MeasurementControl(name='MC')

    test_MC.soft_avg(5)
    snap = dh.snapshot()
    assert snap['instruments'].keys() == {'MC'}
    assert snap['instruments']['MC']['parameters']['soft_avg']['value'] == 5

    test_MC.close()


def test_dynamic_dataset():
    x = ManualParameter('x', unit='m', label='X position')
    y = ManualParameter('y', unit='m', label='Y position')
    z = ManualParameter('z', unit='V', label='Signal amplitude')
    settables = [x, y]
    gettables = [z]
    dset = dh.initialize_dataset(settables, np.empty((8, len(settables))), gettables)

    x0_vals = np.random.random(8)
    x1_vals = np.random.random(8)
    y0_vals = np.random.random(8)
    dset['x0'].values[:] = x0_vals
    dset['x1'].values[:] = x1_vals
    dset['y0'].values[:] = y0_vals

    dset = dh.grow_dataset(dset)
    assert len(dset['x0']) == len(dset['x1']) == len(dset['y0']) == 16
    assert np.isnan(dset['x0'][8:]).all()
    assert np.isnan(dset['x1'][8:]).all()
    assert np.isnan(dset['y0'][8:]).all()

    x0_vals_ext = np.random.random(3)
    x1_vals_ext = np.random.random(3)
    y0_vals_ext = np.random.random(3)

    dset['x0'].values[8:11] = x0_vals_ext
    dset['x1'].values[8:11] = x1_vals_ext
    dset['y0'].values[8:11] = y0_vals_ext

    dset = dh.trim_dataset(dset)
    assert len(dset['x0']) == len(dset['x1']) == len(dset['y0']) == 11
    np.array_equal(dset['x0'], np.concatenate([x0_vals, x0_vals_ext]))
    np.array_equal(dset['x1'], np.concatenate([x1_vals, x1_vals_ext]))
    np.array_equal(dset['y0'], np.concatenate([y0_vals, y0_vals_ext]))

    assert not np.isnan(dset['x0']).any()
    assert not np.isnan(dset['x1']).any()
    assert not np.isnan(dset['y0']).any()



import numpy as np
from quantify.visualization import PlotMonitor_pyqt

from tests.helpers import get_test_data_dir
from quantify.data.types import TUID
from quantify.data.handling import set_datadir


test_datadir = get_test_data_dir()


class TestPlotMonitor_pyqt:

    @classmethod
    def setup_class(cls):
        cls.plotmon = PlotMonitor_pyqt(name='plotmon')
        # ensures the default datadir is used which is excluded from git
        set_datadir(test_datadir)

    @classmethod
    def teardown_class(cls):
        cls.plotmon.close()
        set_datadir(None)

    def test_attributes_created_during_init(self):
        hasattr(self.plotmon, 'main_QtPlot')
        hasattr(self.plotmon, 'secondary_QtPlot')

    def test_validator_accepts_TUID_objects(self):
        self.plotmon.tuid(TUID('20200430-170837-001-315f36'))

    def test_basic_1D_plot(self):
        # Test 1D plotting using an example dataset
        self.plotmon.tuid('20200430-170837-001-315f36')
        self.plotmon.update()

        x = self.plotmon.curves[0]['config']['x']
        y = self.plotmon.curves[0]['config']['y']

        x_exp = np.linspace(0, 5, 50)
        y_exp = np.cos(np.pi*x_exp)
        np.testing.assert_allclose(x, x_exp)
        np.testing.assert_allclose(y, y_exp)

    def test_basic_2D_plot(self):
        # Test 1D plotting using an example dataset
        self.plotmon.tuid('20200504-191556-002-4209ee')
        self.plotmon.update()

        x = self.plotmon.curves[0]['config']['x']
        y = self.plotmon.curves[0]['config']['y']

        x_exp = np.linspace(0, 5, 50)
        y_exp = np.cos(np.pi*x_exp)*-1
        np.testing.assert_allclose(x[:50], x_exp)
        np.testing.assert_allclose(y[:50], y_exp)

        cfg = self.plotmon.secondary_QtPlot.traces[0]['config']
        assert np.shape(cfg['z']) == (11, 50)
        assert cfg['xlabel'] == 'Time'
        assert cfg['xunit'] == 's'
        assert cfg['ylabel'] == 'Amplitude'
        assert cfg['yunit'] == 'V'
        assert cfg['zlabel'] == 'Signal level'
        assert cfg['zunit'] == 'V'



from quantify.utilities.general import delete_keys_from_dict, make_hash


def test_delete_keys_from_dict():

    test_dict = {"a": 5, "b": 6, "c": {"D": 4, "E": 8}}

    assert "a" in test_dict.keys()
    test_dict = delete_keys_from_dict(test_dict, {"a"})
    assert "a" not in test_dict.keys()

    assert "D" in test_dict['c'].keys()
    test_dict = delete_keys_from_dict(test_dict, {"D"})
    assert "D" not in test_dict['c'].keys()


def test_make_hash():

    my_test_dict = {'a': 5, "nested_dict": {
        "a": 2, "c": 4, "B": 'str'}, 'b': 24}

    same_test_dict_diff_order = {'a': 5, 'b': 24, "nested_dict": {
        "a": 2, "c": 4, "B": 'str'}, }

    diff_test_dict = {"nested_dict": {
        "a": 2, "c": 4, "B": 'str'}, 'b': 24}

    test_hash = make_hash(my_test_dict)
    same_test_hash = make_hash(same_test_dict_diff_order)

    assert test_hash == same_test_hash

    diff_hash = make_hash(diff_test_dict)

    assert test_hash != diff_hash

    # modify dict in place, the object id won't change
    my_test_dict['efg'] = 15
    new_hash = make_hash(my_test_dict)
    assert test_hash != new_hash





import pytest
from datetime import datetime
from quantify.data.types import TUID


def test_TUID():
    tuid = TUID('20200409-123015-123-abcdef')

    dt = tuid.datetime()
    assert isinstance(dt, datetime)
    assert isinstance(tuid, str)

    assert dt.year == 2020
    assert dt.month == 4
    assert dt.day == 9

    assert dt.hour == 12
    assert dt.minute == 30
    assert dt.second == 15

    assert tuid.uuid() == 'abcdef'

    with pytest.raises(ValueError):
        tuid = TUID('200409-123015-123-abcdef')

    with pytest.raises(ValueError):
        tuid = TUID('200409-123015-123-abcdefasf')


def test_TUID_validation():
    TUID.is_valid('20200409-123015-123-abcdef')

    problems = [
        '20200409-123015-123-a',  # too short uid
        '20200409-123015-123-a135bcdefasf',  # too long uid
        '20200409-123015-abcdef',  # missing milliseconds
        '200409-123015-123-abcdef',  # 2 digit year
        '20200409123015-123-abcdef',  # missing dash
        "20200924-152319a414-131ece",  # wrong separator character
        "20200924-959399-414-131ece",  # impossible time
        "20200961-152319-414-131ece",  # impossible date
    ]

    for case in problems:
        with pytest.raises(ValueError):
            TUID.is_valid(case)



import time
import random
import xarray as xr
import numpy as np
import pickle
import pytest
import adaptive
import os
from threading import Timer
from scipy import optimize
from qcodes import ManualParameter, Parameter
from qcodes.instrument.base import Instrument
from qcodes.utils import validators as vals
from quantify.measurement.control import MeasurementControl, tile_setpoints_grid
from quantify.data.handling import set_datadir
from quantify.data.types import TUID
from quantify.visualization.pyqt_plotmon import PlotMonitor_pyqt
from quantify.utilities.experiment_helpers import load_settings_onto_instrument
from tests.helpers import get_test_data_dir


test_datadir = get_test_data_dir()


def CosFunc(t, amplitude, frequency, phase):
    """A simple cosine function"""
    return amplitude * np.cos(2 * np.pi * frequency * t + phase)


def SinFunc(t, amplitude, frequency, phase):
    return amplitude * np.sin(2 * np.pi * frequency * t + phase)


# Parameters are created to emulate a system being measured
t = ManualParameter('t', initial_value=1, unit='s', label='Time')
amp = ManualParameter('amp', initial_value=1, unit='V', label='Amplitude')
freq = ManualParameter('freq', initial_value=1, unit='Hz', label='Frequency')


def sine_model():
    return SinFunc(t(), amplitude=amp(), frequency=freq(), phase=0)


def cosine_model():
    return CosFunc(t(), amplitude=amp(), frequency=freq(), phase=0)


# We wrap our function in a Parameter to be able to give
sig = Parameter(name='sig', label='Signal level', unit='V', get_cmd=cosine_model)


class DualWave:
    def __init__(self):
        self.name = ['sin', 'cos']
        self.unit = ['V', 'V']
        self.label = ['Sine', 'Cosine']

    def get(self):
        return np.array([sine_model(), cosine_model()])


class DummyParHolder(Instrument):
    def __init__(self, name):
        super().__init__(name)

        for parname in ["x", "y", "z", "x0", "y0", "z0"]:
            self.add_parameter(
                parname,
                unit="m",
                parameter_class=ManualParameter,
                vals=vals.Numbers(),
                initial_value=0.,
            )

        self.add_parameter(
            "noise",
            unit="V",
            label="white noise amplitude",
            parameter_class=ManualParameter,
            vals=vals.Numbers(),
            initial_value=0,
        )

        self.add_parameter(
            "delay",
            unit="s",
            label="Sampling delay",
            parameter_class=ManualParameter,
            vals=vals.Numbers(),
            initial_value=0,
        )

        self.add_parameter("parabola", unit="V", get_cmd=self._measure_parabola)

    def _measure_parabola(self):
        time.sleep(self.delay())
        return (
            (self.x() - self.x0()) ** 2
            + (self.y() - self.y0()) ** 2
            + (self.z() - self.z0()) ** 2
            + self.noise() * np.random.rand(1)
        )


def hardware_mock_values(setpoints):
    return np.sin(setpoints / np.pi)


class DummyHardwareSettable:

    def __init__(self):
        self.name = 'DummyHardwareSettable'
        self.label = 'Amp'
        self.unit = 'V'
        self.soft = False
        self.setpoints = []

    # copy what qcodes does for easy testing interop
    def __call__(self):
        return self.setpoints

    def set(self, setpoints):
        self.setpoints = setpoints


class DummyHardwareGettable:

    def __init__(self, settables, noise=0.0):
        self.name = ['DummyHardwareGettable_0']
        self.unit = ['W']
        self.label = ['Watts']
        self.soft = False
        self.settables = [settables] if not isinstance(settables, list) else settables
        self.noise = noise
        self.get_func = hardware_mock_values

    def set_return_2D(self):
        self.name.append('DummyHardwareGettable_1')
        self.unit.append('V')
        self.label.append('Amp')

    def prepare(self):
        assert len(self.settables) > 0
        for settable in self.settables:
            assert settable is not None

    def _get_data(self):
        return np.array([i() for i in self.settables])

    def get(self):
        data = self._get_data()
        data = self.get_func(data)
        noise = self.noise * (np.random.rand(1, data.shape[1]) - .5)
        data += noise
        return data[0:len(self.name), :]


class TestMeasurementControl:

    @classmethod
    def setup_class(cls):
        cls.MC = MeasurementControl(name='MC')
        # ensures the default datadir is used which is excluded from git
        cls.dummy_parabola = DummyParHolder('parabola')
        set_datadir(None)

    @classmethod
    def teardown_class(cls):
        cls.MC.close()
        cls.dummy_parabola.close()
        set_datadir(None)

    def test_MeasurementControl_name(self):
        assert self.MC.name == 'MC'

    def test_setpoints(self):
        x = np.linspace(0, 10, 11)
        self.MC.setpoints(x)
        assert np.array_equal(self.MC._setpoints[:, 0], x)

        x = np.random.rand(15, 2)
        self.MC.setpoints(x)
        assert np.array_equal(self.MC._setpoints, x)

        x = np.random.rand(15, 4)
        self.MC.setpoints(x)
        assert np.array_equal(self.MC._setpoints, x)

    def test_soft_sweep_1D(self):
        xvals = np.linspace(0, 2*np.pi, 31)

        self.MC.settables(t)
        self.MC.setpoints(xvals)
        self.MC.gettables(sig)
        dset = self.MC.run()

        assert TUID.is_valid(dset.attrs['tuid'])

        expected_vals = CosFunc(t=xvals, amplitude=1, frequency=1, phase=0)

        assert (np.array_equal(dset['x0'].values, xvals))
        assert (np.array_equal(dset['y0'].values, expected_vals))

        assert isinstance(dset, xr.Dataset)
        assert dset.keys() == {'x0', 'y0'}
        assert (np.array_equal(dset['x0'], xvals))
        assert dset['x0'].attrs == {'name': 't', 'long_name': 'Time', 'unit': 's'}
        assert dset['y0'].attrs == {'name': 'sig', 'long_name': 'Signal level', 'unit': 'V'}

    def test_soft_sweep_1D_multi_return(self):
        xvals = np.linspace(0, 2*np.pi, 31)

        self.MC.settables(t)
        self.MC.setpoints(xvals)
        self.MC.gettables(DualWave())
        dset = self.MC.run()

        exp_y0 = SinFunc(xvals, 1, 1, 0)
        exp_y1 = CosFunc(xvals, 1, 1, 0)

        assert dset.keys() == {'x0', 'y0', 'y1'}
        np.testing.assert_array_equal(dset['y0'], exp_y0)
        np.testing.assert_array_equal(dset['y1'], exp_y1)

    def test_soft_averages_soft_sweep_1D(self):
        def rand():
            return random.uniform(0.0, t())
        rand_get = Parameter(name='sig', label='Signal level', unit='V', get_cmd=rand)
        setpoints = np.arange(100.0)
        self.MC.settables(t)
        self.MC.setpoints(setpoints)
        self.MC.gettables(rand_get)
        r_dset = self.MC.run('random')

        self.MC.soft_avg(50)
        avg_dset = self.MC.run('averaged')

        expected_vals = 0.5 * np.arange(100.0)
        r_delta = abs(r_dset['y0'].values - expected_vals)
        avg_delta = abs(avg_dset['y0'].values - expected_vals)
        assert np.mean(avg_delta) < np.mean(r_delta)

    def test_hard_sweep_1D(self):
        x = np.linspace(0, 10, 5)
        device = DummyHardwareSettable()
        self.MC.settables(device)
        self.MC.setpoints(x)
        self.MC.gettables(DummyHardwareGettable(device))
        dset = self.MC.run()

        expected_vals = hardware_mock_values(x)
        assert (np.array_equal(dset['x0'].values, x))
        assert (np.array_equal(dset['y0'].values, expected_vals))

        assert isinstance(dset, xr.Dataset)
        assert dset.keys() == {'x0', 'y0'}
        assert dset['x0'].attrs == {'name': 'DummyHardwareSettable', 'long_name': 'Amp', 'unit': 'V'}
        assert dset['y0'].attrs == {'name': 'DummyHardwareGettable_0', 'long_name': 'Watts', 'unit': 'W'}

    def test_soft_averages_hard_sweep_1D(self):
        setpoints = np.arange(50.0)
        settable = DummyHardwareSettable()
        gettable = DummyHardwareGettable(settable)
        gettable.noise = 0.4
        self.MC.settables(settable)
        self.MC.setpoints(setpoints)
        self.MC.gettables(gettable)
        noisy_dset = self.MC.run('noisy')
        xn_0 = noisy_dset['x0'].values
        expected_vals = hardware_mock_values(xn_0)
        yn_0 = abs(noisy_dset['y0'].values - expected_vals)

        self.MC.soft_avg(5000)
        avg_dset = self.MC.run('averaged')
        yavg_0 = abs(avg_dset['y0'].values - expected_vals)

        np.testing.assert_array_equal(xn_0, setpoints)
        assert np.mean(yn_0) > np.mean(yavg_0)
        np.testing.assert_array_almost_equal(yavg_0, np.zeros(len(xn_0)), decimal=2)

    def test_soft_set_hard_get_1D(self):
        gettable = DummyHardwareGettable(t)

        def mock_get():
            return np.sin(t())

        gettable.get = mock_get
        setpoints = np.linspace(0, 360, 8)
        self.MC.settables(t)
        self.MC.setpoints(setpoints)
        self.MC.gettables(gettable)
        dset = self.MC.run("soft_sweep_hard_det")

        x = dset['x0'].values
        y0 = dset['y0'].values
        np.testing.assert_array_equal(x, setpoints)
        np.testing.assert_array_equal(y0, np.sin(setpoints))

    def test_soft_sweep_2D_grid(self):

        times = np.linspace(0, 5, 20)
        amps = np.linspace(-1, 1, 5)

        self.MC.settables([t, amp])
        self.MC.setpoints_grid([times, amps])

        exp_sp = tile_setpoints_grid([times, amps])
        assert (np.array_equal(self.MC._setpoints, exp_sp))

        self.MC.gettables(sig)
        dset = self.MC.run()

        assert TUID.is_valid(dset.attrs['tuid'])

        expected_vals = CosFunc(t=exp_sp[:, 0], amplitude=exp_sp[:, 1], frequency=1, phase=0)

        assert (np.array_equal(dset['x0'].values, exp_sp[:, 0]))
        assert (np.array_equal(dset['x1'].values, exp_sp[:, 1]))
        assert (np.array_equal(dset['y0'].values, expected_vals))

        # Test properties of the dataset
        assert isinstance(dset, xr.Dataset)
        assert dset.keys() == {'x0', 'x1', 'y0'}

        assert (all(e in dset['x0'].values for e in times))
        assert (all(e in dset['x1'].values for e in amps))

        assert dset['x0'].attrs == {'name': 't', 'long_name': 'Time', 'unit': 's'}
        assert dset['x1'].attrs == {'name': 'amp', 'long_name': 'Amplitude', 'unit': 'V'}
        assert dset['y0'].attrs == {'name': 'sig', 'long_name': 'Signal level', 'unit': 'V'}

    def test_soft_sweep_2D_arbitrary(self):

        r = np.linspace(0, 1.5, 50)
        dt = np.linspace(0, 1, 50)

        f = 10

        # create a fancy polar coordinates loop
        theta = np.cos(2*np.pi*f*dt)

        def polar_coords(r, theta):

            x = r*np.cos(2*np.pi*theta)
            y = r*np.sin(2*np.pi*theta)
            return x, y

        x, y = polar_coords(r, theta)
        setpoints = np.column_stack([x, y])

        self.MC.settables([t, amp])
        self.MC.setpoints(setpoints)
        self.MC.gettables(sig)
        dset = self.MC.run()

        assert TUID.is_valid(dset.attrs['tuid'])

        expected_vals = CosFunc(t=x, amplitude=y, frequency=1, phase=0)

        assert (np.array_equal(dset['x0'].values, x))
        assert (np.array_equal(dset['x1'].values, y))
        assert (np.array_equal(dset['y0'].values, expected_vals))

    def test_hard_sweep_2D_grid(self):
        times = np.linspace(10, 20, 3)
        amps = np.linspace(0, 10, 5)

        settables = [DummyHardwareSettable(), DummyHardwareSettable()]
        gettable = DummyHardwareGettable(settables)
        self.MC.settables(settables)
        self.MC.setpoints_grid([times, amps])
        self.MC.gettables(gettable)
        dset = self.MC.run('2D Hard')

        exp_sp = tile_setpoints_grid([times, amps])
        assert np.array_equal(exp_sp, self.MC._setpoints)
        assert np.array_equal(dset['x0'].values, exp_sp[:, 0])
        assert np.array_equal(dset['x1'].values, exp_sp[:, 1])

        expected_vals = hardware_mock_values(dset['x0'].values)
        assert np.array_equal(dset['y0'].values, expected_vals)

        # Test properties of the dataset
        assert isinstance(dset, xr.Dataset)

        assert dset['x0'].attrs == {'name': 'DummyHardwareSettable', 'long_name': 'Amp', 'unit': 'V'}
        assert dset['x1'].attrs == {'name': 'DummyHardwareSettable', 'long_name': 'Amp', 'unit': 'V'}
        assert dset['y0'].attrs == {'name': 'DummyHardwareGettable_0', 'long_name': 'Watts', 'unit': 'W'}

    def test_hard_sweep_2D_grid_multi_return(self):
        times = np.linspace(10, 20, 3)
        amps = np.linspace(0, 10, 5)

        settables = [DummyHardwareSettable(), DummyHardwareSettable()]
        gettable = DummyHardwareGettable(settables)
        gettable.set_return_2D()
        self.MC.settables(settables)
        self.MC.setpoints_grid([times, amps])
        self.MC.gettables(gettable)
        dset = self.MC.run('2D Hard')

        exp_sp = tile_setpoints_grid([times, amps])
        assert np.array_equal(exp_sp, self.MC._setpoints)
        assert np.array_equal(dset['x0'].values, exp_sp[:, 0])
        assert np.array_equal(dset['x1'].values, exp_sp[:, 1])

        expected_vals = hardware_mock_values(np.stack((dset['x0'].values, dset['x1'].values)))
        assert np.array_equal(dset['y0'].values, expected_vals[0])
        assert np.array_equal(dset['y1'].values, expected_vals[1])

    def test_hard_sweep_2D_grid_multi_return_soft_avg(self):
        x0 = np.arange(5)
        x1 = np.linspace(5, 10, 5)
        settables = [DummyHardwareSettable(), DummyHardwareSettable()]
        gettable = DummyHardwareGettable(settables)
        gettable.noise = 0.4
        gettable.set_return_2D()
        self.MC.settables(settables)
        self.MC.setpoints_grid([x0, x1])
        self.MC.gettables(gettable)
        noisy_dset = self.MC.run('noisy_hard_grid')

        expected_vals = hardware_mock_values(np.stack((noisy_dset['x0'].values, noisy_dset['x1'])))
        yn_0 = abs(noisy_dset['y0'].values - expected_vals[0])
        yn_1 = abs(noisy_dset['y1'].values - expected_vals[1])

        self.MC.soft_avg(1000)
        avg_dset = self.MC.run('avg_hard_grid')
        yavg_0 = abs(avg_dset['y0'].values - expected_vals[0])
        yavg_1 = abs(avg_dset['y1'].values - expected_vals[1])

        assert np.mean(yavg_0) < np.mean(yn_0)
        assert np.mean(yavg_1) < np.mean(yn_1)
        np.testing.assert_array_almost_equal(yavg_0, np.zeros(len(noisy_dset['x0'].values)), decimal=2)
        np.testing.assert_array_almost_equal(yavg_1, np.zeros(len(noisy_dset['x0'].values)), decimal=2)

    def test_hard_sweep_2D_arbitrary(self):
        r = np.linspace(0, 1.5, 5)
        dt = np.linspace(0, 1, 5)
        f = 10
        theta = np.cos(2*np.pi*f*dt)

        def polar_coords(r, theta):
            x = r*np.cos(2*np.pi*theta)
            y = r*np.sin(2*np.pi*theta)
            return x, y

        x, y = polar_coords(r, theta)
        setpoints = np.column_stack([x, y])

        settables = [DummyHardwareSettable(), DummyHardwareSettable()]
        self.MC.settables(settables)
        self.MC.setpoints(setpoints)
        self.MC.gettables(DummyHardwareGettable(settables))
        dset = self.MC.run()

        assert TUID.is_valid(dset.attrs['tuid'])

        expected_vals = hardware_mock_values(x)

        assert np.array_equal(dset['x0'].values, x)
        assert np.array_equal(dset['x1'].values, y)
        assert np.array_equal(dset['y0'].values, expected_vals)

    def test_variable_return_hard_sweep(self):
        counter_param = ManualParameter("counter", initial_value=0)

        def v_size(setpoints):
            idx = counter_param() % 3
            counter_param(counter_param() + 1)
            if idx == 0:
                return 2 * setpoints[:7]
            elif idx == 1:
                return 2 * setpoints[:4]
            elif idx == 2:
                return 2 * setpoints[:]

        setpoints = np.arange(30.0)
        settable = DummyHardwareSettable()
        gettable = DummyHardwareGettable(settable)
        gettable.get_func = v_size
        self.MC.settables(settable)
        self.MC.setpoints(setpoints)
        self.MC.gettables(gettable)
        dset = self.MC.run('varying')

        assert np.array_equal(dset['x0'], setpoints)
        assert np.array_equal(dset['y0'], 2 * setpoints)

    def test_variable_return_hard_sweep_soft_avg(self):
        counter_param = ManualParameter("counter", initial_value=0)

        def v_size(setpoints):
            idx = counter_param() % 3
            counter_param(counter_param() + 1)
            if idx == 0:
                return 2 * setpoints[:7]
            elif idx == 1:
                return 2 * setpoints[:4]
            elif idx == 2:
                return 2 * setpoints[:]

        setpoints = np.arange(30.0)
        settable = DummyHardwareSettable()
        gettable = DummyHardwareGettable(settable)
        gettable.get_func = v_size
        gettable.noise = 0.25
        self.MC.settables(settable)
        self.MC.setpoints(setpoints)
        self.MC.gettables(gettable)
        self.MC.soft_avg(5000)
        dset = self.MC.run('varying')

        assert np.array_equal(dset['x0'], setpoints)
        np.testing.assert_array_almost_equal(dset.y0, 2 * setpoints, decimal=2)

    def test_soft_sweep_3D_grid(self):
        times = np.linspace(0, 5, 2)
        amps = np.linspace(-1, 1, 3)
        freqs = np.linspace(41000, 82000, 2)

        self.MC.settables([t, amp, freq])
        self.MC.setpoints_grid([times, amps, freqs])

        exp_sp = tile_setpoints_grid([times, amps, freqs])
        assert (np.array_equal(self.MC._setpoints, exp_sp))

        self.MC.gettables(sig)
        dset = self.MC.run()

        assert TUID.is_valid(dset.attrs['tuid'])

        expected_vals = CosFunc(t=exp_sp[:, 0], amplitude=exp_sp[:, 1], frequency=exp_sp[:, 2], phase=0)

        assert (np.array_equal(dset['x0'].values, exp_sp[:, 0]))
        assert (np.array_equal(dset['x1'].values, exp_sp[:, 1]))
        assert (np.array_equal(dset['x2'].values, exp_sp[:, 2]))
        assert (np.array_equal(dset['y0'].values, expected_vals))

        # Test properties of the dataset
        assert isinstance(dset, xr.Dataset)
        assert dset.keys() == {'x0', 'x1', 'x2', 'y0'}
        assert (all(e in dset['x0'] for e in times))
        assert (all(e in dset['x1'] for e in amps))
        assert (all(e in dset['x2'] for e in freqs))

        assert dset['x0'].attrs == {'name': 't', 'long_name': 'Time', 'unit': 's'}
        assert dset['x2'].attrs == {'name': 'freq', 'long_name': 'Frequency', 'unit': 'Hz'}
        assert dset['y0'].attrs == {'name': 'sig', 'long_name': 'Signal level', 'unit': 'V'}

    def test_soft_sweep_3D_multi_return(self):
        times = np.linspace(0, 5, 2)
        amps = np.linspace(-1, 1, 3)
        freqs = np.linspace(41000, 82000, 2)

        self.MC.settables([t, amp, freq])
        self.MC.setpoints_grid([times, amps, freqs])
        self.MC.gettables([DualWave(), sig, DualWave()])
        dset = self.MC.run()

        exp_sp = tile_setpoints_grid([times, amps, freqs])
        exp_y0 = exp_y3 = SinFunc(t=exp_sp[:, 0], amplitude=exp_sp[:, 1], frequency=exp_sp[:, 2], phase=0)
        exp_y1 = exp_y2 = exp_y4 = CosFunc(t=exp_sp[:, 0], amplitude=exp_sp[:, 1], frequency=exp_sp[:, 2], phase=0)

        np.testing.assert_array_equal(dset['y0'], exp_y0)
        np.testing.assert_array_equal(dset['y1'], exp_y1)
        np.testing.assert_array_equal(dset['y2'], exp_y2)
        np.testing.assert_array_equal(dset['y3'], exp_y3)
        np.testing.assert_array_equal(dset['y4'], exp_y4)

    def test_hard_sweep_3D_multi_return_soft_averaging(self):
        def v_get(setpoints):
            return setpoints

        times = np.linspace(0, 5, 2)
        amps = np.linspace(-1, 1, 3)
        freqs = np.linspace(41000, 82000, 2)
        hardware_settable_0 = DummyHardwareSettable()
        hardware_settable_1 = DummyHardwareSettable()
        nd_gettable = DummyHardwareGettable([hardware_settable_0, hardware_settable_1])
        nd_gettable.set_return_2D()
        noisy_gettable = DummyHardwareGettable([t])
        noisy_gettable.get_func = v_get
        noisy_gettable.noise = 0.25

        self.MC.settables([t, hardware_settable_0, hardware_settable_1])
        self.MC.setpoints_grid([times, amps, freqs])
        self.MC.gettables([noisy_gettable, nd_gettable])
        self.MC.soft_avg(5000)
        dset = self.MC.run()

        exp_sp = tile_setpoints_grid([times, amps, freqs])
        np.testing.assert_array_almost_equal(dset.y0, exp_sp[:, 0], decimal=2)
        np.testing.assert_array_almost_equal(dset.y1, hardware_mock_values(exp_sp[:, 1]), decimal=4)
        np.testing.assert_array_almost_equal(dset.y2, hardware_mock_values(exp_sp[:, 2]), decimal=4)

    def test_adaptive_no_averaging(self):
        self.MC.soft_avg(5)
        with pytest.raises(ValueError, match=r"software averaging not allowed in adaptive loops; currently set to 5"):
            self.MC.run_adaptive('fail', {})
        self.MC.soft_avg(1)

    def test_adaptive_nelder_mead(self):
        self.MC.settables([self.dummy_parabola.x, self.dummy_parabola.y])
        af_pars = {
            "adaptive_function": optimize.minimize,
            "x0": [-50, -50],
            "method": "Nelder-Mead"
        }
        self.dummy_parabola.noise(0.5)
        self.MC.gettables(self.dummy_parabola.parabola)
        dset = self.MC.run_adaptive('nelder_mead', af_pars)

        assert dset['x0'][-1] < 0.7
        assert dset['x1'][-1] < 0.7
        assert dset['y0'][-1] < 0.7

    def test_adaptive_multi_return(self):
        freq = ManualParameter(name='frequency', unit='Hz', label='Frequency')
        amp = ManualParameter(name='amp', unit='V', label='Amplitude', initial_value=1)
        dummy = ManualParameter(name='dummy', unit='V', label='Dummy', initial_value=99)
        fwhm = 300
        resonance_freq = random.uniform(6e9, 7e9)

        def lorenz():
            return amp() * ((fwhm / 2.) ** 2) / ((freq() - resonance_freq) ** 2 + (fwhm / 2.) ** 2)

        # gimmick class for return
        class ResonancePlus:
            def __init__(self):
                self.name = ['resonance', 'dummy']
                self.label = ['Amplitude', 'Dummy']
                self.unit = ['V', 'V']

            def get(self):
                return [lorenz(), dummy()]

        self.MC.settables(freq)
        af_pars = {
            "adaptive_function": adaptive.learner.Learner1D,
            "goal": lambda l: l.npoints > 5,
            "bounds": (6e9, 7e9),
        }
        self.MC.gettables([ResonancePlus(), dummy])
        dset = self.MC.run_adaptive('multi return', af_pars)
        assert (dset['y1'].values == 99).all()
        assert (dset['y2'].values == 99).all()

    def test_adaptive_bounds_1D(self):
        freq = ManualParameter(name='frequency', unit='Hz', label='Frequency')
        amp = ManualParameter(name='amp', unit='V', label='Amplitude', initial_value=1)
        fwhm = 300
        resonance_freq = random.uniform(6e9, 7e9)

        def lorenz():
            return amp() * ((fwhm / 2.) ** 2) / ((freq() - resonance_freq) ** 2 + (fwhm / 2.) ** 2)

        resonance = Parameter('resonance', unit='V', label='Amplitude', get_cmd=lorenz)

        self.MC.settables(freq)
        af_pars = {
            "adaptive_function": adaptive.learner.Learner1D,
            "goal": lambda l: l.npoints > 20 * 20,
            "bounds": (6e9, 7e9),
        }
        self.MC.gettables(resonance)
        dset = self.MC.run_adaptive('adaptive sample', af_pars)
        print('jej')

    def test_adaptive_sampling(self):
        self.dummy_parabola.noise(0)
        self.MC.settables([self.dummy_parabola.x, self.dummy_parabola.y])
        af_pars = {
            "adaptive_function": adaptive.learner.Learner2D,
            "goal": lambda l: l.npoints > 20 * 20,
            "bounds": ((-50, 50), (-20, 30)),
        }
        self.MC.gettables(self.dummy_parabola.parabola)
        dset = self.MC.run_adaptive('adaptive sample', af_pars)
        # todo pycqed has no verification step here, what should we do?

    def test_adaptive_skoptlearner(self):
        self.dummy_parabola.noise(0)
        self.MC.settables([self.dummy_parabola.x, self.dummy_parabola.y])
        af_pars = {
            "adaptive_function": adaptive.SKOptLearner,
            "goal": lambda l: l.npoints > 30,
            "dimensions": [(-50.0, +50.0), (-20.0, +30.0)],
            "base_estimator": "gp",
            "acq_func": "EI",
            "acq_optimizer": "lbfgs",
        }
        self.MC.gettables(self.dummy_parabola.parabola)
        dset = self.MC.run_adaptive('skopt', af_pars)
        # todo pycqed has no verification step here, what should we do?

    def test_progress_callback(self):
        progress_param = ManualParameter("progress", initial_value=0)

        def set_progress_param_callable(progress):
            progress_param(progress)

        self.MC.on_progress_callback(set_progress_param_callable)
        assert progress_param() == 0

        xvals = np.linspace(0, 2*np.pi, 31)
        self.MC.settables(t)
        self.MC.setpoints(xvals)
        self.MC.gettables(sig)
        self.MC.run()

        assert progress_param() == 100

    def test_MC_plotmon_integration(self):
        plotmon = PlotMonitor_pyqt('plotmon_MC')

        self.MC.instr_plotmon(plotmon.name)

        assert plotmon.tuid() == 'latest'

        times = np.linspace(0, 5, 18)
        amps = np.linspace(-1, 1, 5)

        self.MC.settables([t, amp])
        self.MC.setpoints_grid([times, amps])
        self.MC.gettables(sig)
        self.MC.run('2D Cosine test')

        assert plotmon.tuid() != 'latest'

        plotmon.close()
        self.MC.instr_plotmon('')

    def test_instrument_settings_from_disk(self):
        load_settings_onto_instrument(self.dummy_parabola, TUID('20200814-134652-492-fbf254'), test_datadir)
        assert self.dummy_parabola.x() == 40.0
        assert self.dummy_parabola.y() == 90.0
        assert self.dummy_parabola.z() == -20.0
        assert self.dummy_parabola.delay() == 10.0

        non_existing = DummyParHolder('the mac')
        with pytest.raises(ValueError, match='Instrument "the mac" not found in snapshot'):
            load_settings_onto_instrument(non_existing, TUID('20200814-134652-492-fbf254'), test_datadir)

        non_existing.close()


def test_tile_setpoints_grid():
    x = np.arange(5)
    y = np.linspace(-1, 1, 3)

    sp = tile_setpoints_grid([x, y])
    assert sp[:, 0].all() == np.tile(np.arange(5), 3).all()
    assert sp[:, 1].all() == np.repeat(y, 5).all()

    z = np.linspace(100, 200, 2)
    sp = tile_setpoints_grid([x, y, z])
    assert (all(e in sp[:, 0] for e in x))
    assert (all(e in sp[:, 1] for e in y))
    assert (all(e in sp[:, 2] for e in z))



# -----------------------------------------------------------------------------
# Description:    Plot interpolations.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import numpy as np
import logging
from scipy import interpolate


def areas(ip):
    p = ip.tri.points[ip.tri.vertices]
    q = p[:, :-1, :] - p[:, -1, None, :]
    areas = abs(q[:, 0, 0] * q[:, 1, 1] - q[:, 0, 1] * q[:, 1, 0]) / 2
    return areas


def scale(points, xy_mean, xy_scale):
    points = np.asarray(points, dtype=float)
    return (points - xy_mean) / xy_scale


def unscale(points, xy_mean, xy_scale):
    points = np.asarray(points, dtype=float)
    return points * xy_scale + xy_mean


def interpolate_heatmap(x, y, z, n: int = None, interp_method: str = "linear"):
    """
    The output of this method can directly be used for plt.imshow(z_grid, extent=extent, aspect='auto')
    where the extent is determined by the min and max of the x_grid and y_grid.

    The output can also be used as input for ax.pcolormesh(x, y, Z,**kw)

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        x data points
    y : :class:`numpy.ndarray`
        y data points
    z : :class:`numpy.ndarray`
        z data points
    n : int
        number of points for each dimension on the interpolated grid if set to None will auto determine amount of
        points needed
    interp_method : str
        one of {"linear", "nearest", "deg"}, determines what interpolation method is used.
    Returns
    -------
    x_grid : :class:`numpy.ndarray`
        N*1 array of x-values of the interpolated grid
    y_grid : :class:`numpy.ndarray`
        N*1 array of x-values of the interpolated grid
    z_grid : :class:`numpy.ndarray`
        N*N array of z-values that form a grid.
    """

    points = list(zip(x, y))
    lbrt = np.min(points, axis=0), np.max(points, axis=0)
    lbrt = lbrt[0][0], lbrt[0][1], lbrt[1][0], lbrt[1][1]

    xy_mean = np.mean([lbrt[0], lbrt[2]]), np.mean([lbrt[1], lbrt[3]])
    xy_scale = np.ptp([lbrt[0], lbrt[2]]), np.ptp([lbrt[1], lbrt[3]])

    # interpolation needs to happen on a rescaled grid, this is somewhat akin
    # to an assumption in the interpolation that the scale of the experiment
    # is chosen sensibly. N.B. even if interp_method == "nearest" the linear
    # interpolation is used to determine the amount of grid points.
    ip = interpolate.LinearNDInterpolator(scale(points, xy_mean=xy_mean, xy_scale=xy_scale), z)

    if n is None:
        # Calculate how many grid points are needed.
        # factor from A=√3/4 * a² (equilateral triangle)
        # N.B. a factor 4 was added as there were to few points for uniform grid otherwise.
        all_areas = areas(ip)
        area_min = all_areas[all_areas > 0.].min()
        n = int(0.658 / np.sqrt(area_min)) * 4
        n = max(n, 10)
        if n > 500:
            logging.debug("n: {} larger than 500".format(n))
            n = 500

    x_lin = y_lin = np.linspace(-0.5, 0.5, n)

    if interp_method == "linear":
        z_grid = ip(x_lin[:, None], y_lin[None, :]).squeeze()
    elif interp_method == "nearest":
        ip = interpolate.NearestNDInterpolator(scale(points, xy_mean=xy_mean, xy_scale=xy_scale), z)
        z_grid = ip(x_lin[:, None], y_lin[None, :]).squeeze()
    elif interp_method == "deg":
        # Circular interpolation in deg units
        phases = np.deg2rad(z)
        newdata_cos = np.cos(phases)
        newdata_sin = np.sin(phases)

        ip_cos = interpolate.LinearNDInterpolator(scale(points, xy_mean=xy_mean, xy_scale=xy_scale), newdata_cos)
        newdata_cos = ip_cos(x_lin[:, None], y_lin[None, :]).squeeze()

        ip_sin = interpolate.LinearNDInterpolator(scale(points, xy_mean=xy_mean, xy_scale=xy_scale), newdata_sin)
        newdata_sin = ip_sin(x_lin[:, None], y_lin[None, :]).squeeze()

        z_grid = (np.rad2deg(np.arctan2(newdata_sin, newdata_cos)) % 360).squeeze()
    else:
        raise ValueError("interp_method must be one of {'linear', 'nearest', 'deg'}")

    # x and y grid points need to be rescaled from the linearly chosen points
    points_grid = unscale(list(zip(x_lin, y_lin)), xy_mean=xy_mean, xy_scale=xy_scale)
    x_grid = points_grid[:, 0]
    y_grid = points_grid[:, 1]

    return x_grid, y_grid, z_grid.T



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

with open('AUTHORS.rst') as authors_file:
    authors = authors_file.read()

with open('requirements_dev.txt') as test_reqs:
    test_requirements = test_reqs.read().splitlines()

requirements = [
    'qcodes',
    'numpy',
    'scipy',
    'xarray',
    'columnar',
    'xxhash',
    'matplotlib',
    'lmfit',
    'pyqt5==5.14.0',
    'pyqtgraph',
    'plotly',
    'jsonschema',
    'adaptive'
]

setup_requirements = ['pytest-runner', ]

setup(
    author="The Quantify consortium consisting of Qblox and Orange Quantum Systems",
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Unified quantum computing, solid-state and pulse sequencing physical experimentation framework.",
    install_requires=requirements,
    license="BSD-4 license",
    long_description=readme + '\n\n' + authors + '\n\n' + history,
    include_package_data=True,
    keywords='quantify-core',
    name='quantify-core',
    packages=find_packages(include=['quantify', 'quantify.*']),
    package_data={'': ['*.json']},  # ensures JSON schema are included
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://https://gitlab.com/quantify-os/quantify-core',
    version='0.2.0',
    zip_safe=False,
)



# -----------------------------------------------------------------------------
# Description:    Utilities for managing SI units with plotting systems.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import matplotlib.pyplot as plt
import matplotlib
import numpy as np

golden_mean = (np.sqrt(5)-1.0)/2.0  # Aesthetic ratio
single_col_figsize = (3.39, golden_mean*3.39)
double_col_figsize = (6.9, golden_mean*6.9)
thesis_col_figsize = (12.2/2.54, golden_mean*12.2/2.54)


def set_xlabel(axis, label, unit=None, **kw):
    """
    Add a unit aware x-label to an axis object.

    Parameters
    ----------
    axis
        matplotlib axis object to set label on
    label
        the desired label
    unit
        the unit
    **kw
        keyword argument to be passed to matplotlib.set_xlabel
    """
    if unit is not None and unit != '':
        xticks = axis.get_xticks()
        scale_factor, unit = SI_prefix_and_scale_factor(val=max(abs(xticks)), unit=unit)
        formatter = matplotlib.ticker.FuncFormatter(lambda x, pos: '{:.4g}'.format(x*scale_factor))

        axis.xaxis.set_major_formatter(formatter)
        axis.set_xlabel(label+' ({})'.format(unit), **kw)
    else:
        axis.set_xlabel(label, **kw)
    return axis


def set_ylabel(axis, label, unit=None, **kw):
    """
    Add a unit aware y-label to an axis object.

    Parameters
    ----------
    axis
        matplotlib axis object to set label on
    label
        the desired label
    unit
        the unit
    **kw
        keyword argument to be passed to matplotlib.set_ylabel
    """
    if unit is not None and unit != '':
        yticks = axis.get_yticks()
        scale_factor, unit = SI_prefix_and_scale_factor(
            val=max(abs(yticks)), unit=unit)
        formatter = matplotlib.ticker.FuncFormatter(
            lambda x, pos: '{:.6g}'.format(x*scale_factor))

        axis.yaxis.set_major_formatter(formatter)

        axis.set_ylabel(label+' ({})'.format(unit), **kw)
    else:
        axis.set_ylabel(label, **kw)
    return axis


def set_cbarlabel(cbar, label, unit=None, **kw):
    """
    Add a unit aware z-label to a colorbar object

    Parameters
    ----------
    cbar
        colorbar object to set label on
    label
        the desired label
    unit
        the unit
    **kw
        keyword argument to be passed to cbar.set_label
    """
    if unit is not None and unit != '':
        zticks = cbar.get_ticks()
        scale_factor, unit = SI_prefix_and_scale_factor(
            val=max(abs(zticks)), unit=unit)
        cbar.set_ticks(zticks)
        cbar.set_ticklabels(zticks*scale_factor)
        cbar.set_label(label + ' ({})'.format(unit))

    else:
        cbar.set_label(label, **kw)
    return cbar


SI_PREFIXES = dict(zip(range(-24, 25, 3), 'yzafpnμm kMGTPEZY'))
SI_PREFIXES[0] = ""

# N.B. not all of these are SI units, however, all of these support SI prefixes
SI_UNITS = 'm,s,g,W,J,V,A,F,T,Hz,Ohm,S,N,C,px,b,B,K,Bar,' \
           'Vpeak,Vpp,Vp,Vrms,$\Phi_0$,A/s'.split(',')  # noqa: W605


def SI_prefix_and_scale_factor(val, unit=None):
    """
    Takes in a value and unit and if applicable returns the proper scale factor and SI prefix.

    Parameters
    ----------
    val : float
        the value
    unit : str
        the unit of the value
    Returns
    -------
    scale_factor : float
        scale_factor needed to convert value
    unit : str
        unit including the prefix
    """
    if unit in SI_UNITS:
        try:
            with np.errstate(all="ignore"):
                prefix_power = np.log10(abs(val))//3 * 3
                prefix = SI_PREFIXES[prefix_power]
                # Greek symbols not supported in tex
                if plt.rcParams['text.usetex'] and prefix == 'μ':
                    prefix = r'$\mu$'

            return 10 ** -prefix_power,  prefix + unit
        except (KeyError, TypeError):
            pass

    return 1, unit if unit is not None else ""


def SI_val_to_msg_str(val: float, unit: str = None, return_type=str):
    """
    Takes in a value  with optional unit and returns a string tuple consisting of (value_str, unit) where the value
    and unit are rescaled according to SI prefixes, IF the unit is an SI unit (according to the comprehensive list of
    SI units in this file ;).

    the value_str is of the type specified in return_type (str) by default.
    """
    sc, new_unit = SI_prefix_and_scale_factor(val, unit)
    try:
        new_val = sc*val
    except TypeError:
        return return_type(val), unit

    return return_type(new_val), new_unit



# -----------------------------------------------------------------------------
# Description:    Module containing the pyqtgraph based plotting monitor.
# Repository:     https://gitlab.com/quantify-os/quantify-core
# Copyright (C) Qblox BV & Orange Quantum Systems Holding BV (2020)
# -----------------------------------------------------------------------------
import numpy as np

from qcodes import validators as vals
from qcodes.instrument.base import Instrument
from qcodes.instrument.parameter import ManualParameter
from qcodes.plots.colors import color_cycle
from qcodes.plots.pyqtgraph import QtPlot, TransformState

from quantify.data.handling import load_dataset, get_latest_tuid
from quantify.visualization.plot_interpolation import interpolate_heatmap


class PlotMonitor_pyqt(Instrument):
    """
    Pyqtgraph based plot monitor.

    A plot monitor is intended to provide a real-time visualization of a dataset.
    """

    def __init__(self, name: str):
        """
        Creates an instance of the Measurement Control.

        Parameters
        ----------
        name : str
            name
        """
        super().__init__(name=name)

        # Paramaters are attributes that we include in logging
        # and intend the user to change.

        self.add_parameter(
            "tuid",
            docstring="The tuid of the dataset to monitor",
            parameter_class=ManualParameter,
            vals=vals.Strings(),
            initial_value='latest',)

        # used to track if tuid changed
        self._last_tuid = None

        self.create_plot_monitor()

        # convenient access to curve variable for updating
        self.curves = []
        self._im_curves = []
        self._im_scatters = []
        self._im_scatters_last = []

    def create_plot_monitor(self):
        """
        Creates the PyQtGraph plotting monitors.
        Can also be used to recreate these when plotting has crashed.
        """
        if hasattr(self, "main_QtPlot"):
            del self.main_QtPlot
        if hasattr(self, "secondary_QtPlot"):
            del self.secondary_QtPlot

        self.main_QtPlot = QtPlot(
            window_title="Main plotmon of {}".format(self.name),
            figsize=(600, 400))
        self.secondary_QtPlot = QtPlot(
            window_title="Secondary plotmon of {}".format(self.name),
            figsize=(600, 400))

    def _initialize_plot_monitor(self, tuid):
        """
        Clears data in plot monitors and sets it up with the data from the dataset
        """

        # Clear the plot monitors if required.
        if self.main_QtPlot.traces:
            self.main_QtPlot.clear()

        if self.secondary_QtPlot.traces:
            self.secondary_QtPlot.clear()

        self.curves = []
        self._im_curves = []
        self._im_scatters = []
        self._im_scatters_last = []

        # TODO add persistence based on previous dataset
        dset = load_dataset(tuid=tuid)

        set_parnames = list(filter(lambda k: 'x' in k, dset.keys()))
        get_parnames = list(filter(lambda k: 'y' in k, dset.keys()))

        #############################################################
        plot_idx = 1
        for yi in get_parnames:
            for xi in set_parnames:
                # TODO: add persist data here.
                self.main_QtPlot.add(
                    x=dset[xi].values, y=dset[yi].values,
                    subplot=plot_idx,
                    xlabel=dset[xi].attrs['long_name'],
                    xunit=dset[xi].attrs['unit'],
                    ylabel=dset[yi].attrs['long_name'],
                    yunit=dset[yi].attrs['unit'],
                    symbol='o', symbolSize=5,
                )
                # Manual counter is used because we may want to add more
                # than one quantity per panel
                plot_idx += 1
                self.curves.append(self.main_QtPlot.traces[-1])
            self.main_QtPlot.win.nextRow()

        #############################################################
        # Add a square heatmap
        if dset.attrs['2D-grid']:
            plot_idx = 1
            for yi in get_parnames:

                cmap = 'viridis'
                zrange = None

                x = dset['x0'].values[:dset.attrs['xlen']]
                y = dset['x1'].values[::dset.attrs['xlen']]
                z = np.reshape(dset[yi].values,
                               (len(x), len(y)), order='F').T
                config_dict = {
                    "x": x,
                    "y": y,
                    "z": z,
                    "xlabel": dset['x0'].attrs['long_name'],
                    "xunit": dset['x0'].attrs['unit'],
                    "ylabel": dset['x1'].attrs['long_name'],
                    "yunit": dset['x1'].attrs['unit'],
                    "zlabel": dset[yi].attrs['long_name'],
                    "zunit": dset[yi].attrs['unit'],
                    "subplot": plot_idx,
                    "cmap": cmap,
                }
                if zrange is not None:
                    config_dict["zrange"] = zrange
                self.secondary_QtPlot.add(**config_dict)
                plot_idx += 1

        #############################################################
        # if data is not on a grid but is 2D it makes sense to interpolate

        elif len(set_parnames) == 2:
            plot_idx = 1
            for yi in get_parnames:

                cmap = 'viridis'
                zrange = None

                config_dict = {
                    "x": [0, 1],
                    "y": [0, 1],
                    "z": np.zeros([2, 2]),
                    "xlabel": dset['x0'].attrs['long_name'],
                    "xunit": dset['x0'].attrs['unit'],
                    "ylabel": dset['x1'].attrs['long_name'],
                    "yunit": dset['x1'].attrs['unit'],
                    "zlabel": dset[yi].attrs['long_name'],
                    "zunit": dset[yi].attrs['unit'],
                    "subplot": plot_idx,
                    "cmap": cmap,
                }
                if zrange is not None:
                    config_dict["zrange"] = zrange
                self.secondary_QtPlot.add(**config_dict)
                self._im_curves.append(self.secondary_QtPlot.traces[-1])

                # used to mark the interpolation points
                self.secondary_QtPlot.add(
                    x=[], y=[],
                    pen=None,
                    color=1.0, width=0, symbol="o", symbolSize=4,
                    subplot=plot_idx,
                    xlabel=dset['x0'].attrs['long_name'],
                    xunit=dset['x0'].attrs['unit'],
                    ylabel=dset['x1'].attrs['long_name'],
                    yunit=dset['x1'].attrs['unit'],)
                self._im_scatters.append(self.secondary_QtPlot.traces[-1])

                # used to mark the last N-interpolation points
                self.secondary_QtPlot.add(
                    x=[], y=[],
                    color=color_cycle[3],  # marks the point red
                    width=0, symbol="o", symbolSize=7,
                    subplot=plot_idx,
                    xlabel=dset['x0'].attrs['long_name'],
                    xunit=dset['x0'].attrs['unit'],
                    ylabel=dset['x1'].attrs['long_name'],
                    yunit=dset['x1'].attrs['unit'],)
                self._im_scatters_last.append(self.secondary_QtPlot.traces[-1])

                plot_idx += 1

    def update(self):
        if self.tuid() == 'latest':
            # this should automatically set tuid to the most recent tuid.
            tuid = get_latest_tuid()
        else:
            tuid = self.tuid()

        # If tuid has changed, we need to initialize the figure
        if tuid != self._last_tuid:
            # need to initialize the plot monitor
            self._initialize_plot_monitor(tuid)
            self._last_tuid = tuid

        # otherwise we simply update
        else:
            dset = load_dataset(tuid=tuid)
            set_parnames = list(filter(lambda k: 'x' in k, dset.keys()))
            get_parnames = list(filter(lambda k: 'y' in k, dset.keys()))
            # Only updates the main monitor currently.

            #############################################################
            i = 0
            for yi in get_parnames:
                for xi in set_parnames:
                    self.curves[i]["config"]["x"] = dset[xi].values
                    self.curves[i]["config"]["y"] = dset[yi].values
                    i += 1
            self.main_QtPlot.update_plot()

            #############################################################
            # Add a square heatmap
            if dset.attrs['2D-grid']:
                for yidx, yi in enumerate(get_parnames):
                    Z = np.reshape(dset[yi].values,
                                   (dset.attrs['xlen'], dset.attrs['ylen']),
                                   order='F').T
                    self.secondary_QtPlot.traces[yidx]['config']['z'] = Z
                self.secondary_QtPlot.update_plot()

            #############################################################
            # if data is not on a grid but is 2D it makes sense to interpolate
            elif len(set_parnames) == 2:
                for yidx, yi in enumerate(get_parnames):
                    # exists to force reset the x- and y-axis scale
                    new_sc = TransformState(0, 1, True)

                    x = dset['x0'].values[~np.isnan(dset)['y0']]
                    y = dset['x1'].values[~np.isnan(dset)['y0']]
                    z = dset[yi].values[~np.isnan(dset)['y0']]
                    # interpolation needs to be meaningful
                    if len(z) < 8:
                        break
                    x_grid, y_grid, z_grid = interpolate_heatmap(x=x, y=y, z=z, interp_method='linear')

                    trace = self._im_curves[yidx]
                    trace['config']['x'] = x_grid
                    trace['config']['y'] = y_grid
                    trace['config']['z'] = z_grid
                    # force rescale axis so marking datapoints works
                    trace["plot_object"]["scales"]["x"] = new_sc
                    trace["plot_object"]["scales"]["y"] = new_sc

                    # Mark all measured points on which the interpolation
                    # is based
                    trace = self._im_scatters[yidx]
                    trace['config']['x'] = x
                    trace['config']['y'] = y

                    trace = self._im_scatters_last[yidx]
                    trace['config']['x'] = x[-5:]
                    trace['config']['y'] = y[-5:]

                self.secondary_QtPlot.update_plot()



import numpy as np
import matplotlib.pyplot as plt
from quantify.visualization.SI_utilities import SI_prefix_and_scale_factor
from quantify.visualization.SI_utilities import set_xlabel, set_ylabel
from quantify.visualization.SI_utilities import SI_val_to_msg_str


def test_non_SI():
    unit = 'arb.unit.'
    scale_factor, post_unit = SI_prefix_and_scale_factor(val=5, unit=unit)
    assert scale_factor == 1
    assert unit == post_unit


def test_SI_scale_factors():
    unit = 'V'
    scale_factor, post_unit = SI_prefix_and_scale_factor(val=5, unit=unit)
    assert scale_factor == 1
    assert ''+unit == post_unit

    scale_factor, post_unit = SI_prefix_and_scale_factor(val=5000,
                                                         unit=unit)
    assert scale_factor, 1 == 1000
    assert 'k'+unit == post_unit

    scale_factor, post_unit = SI_prefix_and_scale_factor(val=0.05,
                                                         unit=unit)
    assert scale_factor == 1000
    assert 'm'+unit == post_unit


def test_label_scaling():
    """
    This test creates a dummy plot and checks if the tick labels are
    rescaled correctly
    """
    f, ax = plt.subplots()
    x = np.linspace(-6, 6, 101)
    y = np.cos(x)
    ax.plot(x*1000, y/1e5)

    set_xlabel(ax, 'Distance', 'm')
    set_ylabel(ax, 'Amplitude', 'V')

    xlab = ax.get_xlabel()
    ylab = ax.get_ylabel()
    assert xlab == 'Distance (km)'
    assert ylab == 'Amplitude (μV)'


def test_SI_val_to_msg_str():
    val, unit = SI_val_to_msg_str(1030, 'm')
    assert val == str(1.03)
    assert unit == 'km'


# class test_plot_lmfit_res(unittest.TestCase):

#     def test_plot_model_result(self):
#         def line(a, b, x):
#             return a*x+b

#         a = .1
#         b = 5
#         x = np.linspace(0, 20, 31)
#         y = line(a, b, x)

#         line_model = lmfit.Model(line, independent_vars='x')
#         line_model.set_param_hint('a', value=a)
#         line_model.set_param_hint('b', value=b)
#         params = line_model.make_params()
#         fit_res = line_model.fit(y, x=x, params=params)

#         f, ax = plt.subplots()
#         plot_lmfit_res(fit_res, ax=ax, plot_kws={'color': 'C1'},
#                        plot_init=True, plot_init_kws={'ls': '--'})
